import { TypeKeys } from "web-utility";
import { HTTPClient, Context } from "koajax";
import { ValidationError } from "class-validator";
export type IDType = number | string;
export type DataObject = Record<string, any>;
export type NewData<T extends DataObject> = Partial<Omit<T, TypeKeys<T, DataObject> | TypeKeys<T, DataObject[]>> & {
    [K in TypeKeys<T, DataObject>]: IDType;
} & {
    [K in TypeKeys<T, DataObject[]>]: IDType[];
}>;
export type RESTClient = Pick<HTTPClient<Context>, 'baseURI' | 'options' | 'request' | 'get' | 'post' | 'put' | 'patch' | 'delete'>;
export abstract class BaseModel {
    downloading: number;
    uploading: number;
    clear(): this;
}
export abstract class BaseListModel<D extends DataObject> extends BaseModel {
    abstract client: RESTClient;
    abstract baseURI: string;
    currentOne: D;
    validity: InvalidMessage<D>;
    static createNested(parentId: IDType): BaseListModel<{}>;
    clear(): this;
    updateOne(data: NewData<D>, id?: IDType): Promise<D>;
    getOne(id: IDType): Promise<D>;
    removeOne(id: IDType): Promise<void>;
}
export function toggle<T extends BaseModel>(property: TypeKeys<T, boolean | number>): (target: any, key: string, meta: PropertyDescriptor) => void;
export function Input(): ParameterDecorator;
export type InvalidMessage<T> = Partial<Record<keyof T, ValidationError['constraints']>>;
export class InvalidError extends Error {
    static from<T>(state: InvalidMessage<T>): InvalidError;
}
export function validate<T extends BaseListModel<{}>>(): (target: any, key: string, meta: PropertyDescriptor) => void;
export interface PageData<D extends DataObject> {
    pageData: D[];
    totalCount?: number;
}
export abstract class ListModel<D extends DataObject, F extends NewData<D> = NewData<D>> extends BaseListModel<D> {
    pageIndex: number;
    pageSize: number;
    filter: F;
    noMore: boolean;
    totalCount: number;
    pageList: D[][];
    get currentPage(): D[];
    get allItems(): any[];
    constructor();
    clear(): this;
    turnTo(pageIndex: number, pageSize?: number): this;
    protected abstract loadPage(pageIndex: number, pageSize: number, filter: F): Promise<PageData<D>>;
    protected loadNewPage(pageIndex: number, pageSize: number, filter: F): Promise<{
        pageData: D[];
        totalCount: number;
    }>;
    protected flushPage(pageIndex: number, pageSize: number, pageData: D[]): D[];
    getList(filter?: F, pageIndex?: number, pageSize?: number): Promise<D[]>;
}
/**
 * Only for Type Hint of {@link BufferListModel} mixin
 *
 * @deprecated
 */
export abstract class BufferList<D extends DataObject, F extends NewData<D> = NewData<D>> extends ListModel<D, F> {
    protected pendingList: ReturnType<ListModel<D, F>['loadPage']>[];
}
export function BufferListModel<D extends DataObject, F extends NewData<D> = NewData<D>>(): abstract new () => BufferList<D, F>;
/**
 * Only for Type Hint of {@link StreamListModel} mixin
 *
 * @deprecated
 */
export abstract class StreamList<D extends DataObject, F extends NewData<D> = NewData<D>> extends ListModel<D, F> {
    baseURI: string;
    protected stream?: AsyncGenerator<D, void, D>;
    protected abstract openStream(filter: F): AsyncGenerator<D, void, D>;
    protected loadPage(pageIndex: number, pageSize: number, filter: F): Promise<{
        pageData: any[];
    }>;
}
export function StreamListModel<D extends DataObject, F extends NewData<D> = NewData<D>>(): abstract new () => StreamList<D, F>;

//# sourceMappingURL=index.d.ts.map
